"""Google earth engine main script."""
import base64
import json

import ee

# import os


class GEE:
    """GEE."""

    def __init__(self, service_account: str, service_key_path: str):
        """Initialize.

        Parameters
        ----------
        service_account: [str]
                service account name
        service_key_path: [str]
                path to the service account json file
        Returns
        -------
        None
        """
        self.initialize(service_account, service_key_path)

        pass

    @staticmethod
    def initialize(service_account: str, service_key: str):
        """Initialize.

            Initialize authenticate and initializes the connection to google earth engine with a service accont file
            content or path

        Parameters
        ----------
        service_account: [str]
                service account name
        service_key: [str]
                path to the service account json file or the content of the service account

        Returns
        -------
        None
        """
        try:
            credentials = ee.ServiceAccountCredentials(service_account, service_key)
        except ValueError:
            credentials = ee.ServiceAccountCredentials(
                service_account, key_data=service_key
            )
        ee.Initialize(credentials=credentials)

    @staticmethod
    def encodeServiceAccount(service_key_dir: str) -> bytes:
        """encodeServiceAccount.

            decodeServiceAccount decode the service account

        Parameters
        ----------
        service_key_dir: [str]

        Returns
        -------
        byte string
        """
        content = json.load(open(service_key_dir))
        dumped_service_account = json.dumps(content)
        encoded_service_account = base64.b64encode(dumped_service_account.encode())
        return encoded_service_account

    @staticmethod
    def decodeServiceAccount(service_key_bytes: bytes) -> str:
        """decodeServiceAccount.

            decodeServiceAccount

        Parameters
        ----------
        service_key_bytes: [bytes]
            the content of the service account encoded with base64

        Returns
        -------
        str:
            google cloud service account content
        """
        service_key = json.loads(base64.b64decode(service_key_bytes).decode())
        return service_key



import datetime as dt
import os
from ftplib import FTP

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from osgeo import gdal
from pyramids.raster import Raster
from pyramids.utils import extractFromGZ
from serapeum_utils.utils import print_progress_bar

from earth2observe.abstractdatasource import AbstractCatalog, AbstractDataSource


class CHIRPS(AbstractDataSource):
    """CHIRPS."""

    api_url: str = "data.chc.ucsb.edu"
    start_date: str = "1981-01-01"
    end_date: str = "Now"
    temporal_resolution = ["daily", "monthly"]
    lat_bondaries = [-50, 50]
    lon_boundaries = [-180, 180]
    globe_fname = "chirps-v2.0"
    clipped_fname = "P_CHIRPS.v2.0"

    def __init__(
        self,
        temporal_resolution: str = "daily",
        start: str = None,
        end: str = None,
        path: str = "",
        variables: list = None,
        lat_lim: list = None,
        lon_lim: list = None,
        fmt: str = "%Y-%m-%d",
    ):
        """CHIRPS.

        Parameters
        ----------
        temporal_resolution (str, optional):
            'daily' or 'monthly'. Defaults to 'daily'.
        start (str, optional):
            [description]. Defaults to ''.
        end (str, optional):
            [description]. Defaults to ''.
        path (str, optional):
            Path where you want to save the downloaded data. Defaults to ''.
        variables (list, optional):
            Variable code: VariablesInfo('day').descriptions.keys(). Defaults to [].
        lat_lim (list, optional):
            [ymin, ymax] (values must be between -50 and 50). Defaults to [].
        lon_lim (list, optional):
            [xmin, xmax] (values must be between -180 and 180). Defaults to [].
        fmt (str, optional):
            [description]. Defaults to "%Y-%m-%d".
        """
        super().__init__(
            start=start,
            end=end,
            variables=variables,
            temporal_resolution=temporal_resolution,
            lat_lim=lat_lim,
            lon_lim=lon_lim,
            fmt=fmt,
            path=path,
        )

    def check_input_dates(
        self, start: str, end: str, temporal_resolution: str, fmt: str
    ):
        """check validity of input dates.

        Parameters
        ----------
        temporal_resolution: (str, optional)
            [description]. Defaults to 'daily'.
        start: (str, optional)
            [description]. Defaults to ''.
        end: (str, optional)
            [description]. Defaults to ''.
        fmt: (str, optional)
            [description]. Defaults to "%Y-%m-%d".
        """
        # check temporal_resolution variables
        if start is None:
            self.start = pd.Timestamp(self.start_date)
        else:
            self.start = dt.datetime.strptime(start, fmt)

        if end is None:
            self.end = pd.Timestamp(self.end_date)
        else:
            self.end = dt.datetime.strptime(end, fmt)

        # Define timestep for the timedates
        if temporal_resolution.lower() == "daily":
            self.time_freq = "D"
            # self.path = os.path.join(path, "precipitation", "chirps", "daily")
        elif temporal_resolution.lower() == "monthly":
            self.time_freq = "MS"
            # self.path = os.path.join(
            #     path, "Precipitation", "CHIRPS", "Monthly"
            # )
        else:
            raise KeyError("The input temporal_resolution interval is not supported")

        # Create days
        self.dates = pd.date_range(self.start, self.end, freq=self.time_freq)

    def initialize(self):
        """Initialize FTP server."""
        print("FTP server datasources does not need server initialization")
        pass

    def create_grid(self, lat_lim: list, lon_lim: list):
        """Create_grid.

            create grid from the lat/lon boundaries

        Parameters
        ----------
        lat_lim: []
            latitude boundaries
        lon_lim: []
            longitude boundaries
        """
        self.lat_lim = []
        self.lon_lim = []
        # Check space variables
        # -50 , 50
        if lat_lim[0] < self.lat_bondaries[0] or lat_lim[1] > self.lat_bondaries[1]:
            print(
                "Latitude above 50N or below 50S is not possible."
                " Value set to maximum"
            )
            self.lat_lim[0] = np.max(lat_lim[0], self.lat_bondaries[0])
            self.lat_lim[1] = np.min(lon_lim[1], self.lat_bondaries[1])
        # -180, 180
        if lon_lim[0] < self.lon_boundaries[0] or lon_lim[1] > self.lon_boundaries[1]:
            print(
                "Longitude must be between 180E and 180W."
                " Now value is set to maximum"
            )
            self.lon_lim[0] = np.max(lat_lim[0], self.lon_boundaries[0])
            self.lon_lim[1] = np.min(lon_lim[1], self.lon_boundaries[1])
        else:
            self.lat_lim = lat_lim
            self.lon_lim = lon_lim

        # Define IDs
        self.yID = 2000 - np.int16(
            np.array(
                [np.ceil((lat_lim[1] + 50) * 20), np.floor((lat_lim[0] + 50) * 20)]
            )
        )
        self.xID = np.int16(
            np.array(
                [np.floor((lon_lim[0] + 180) * 20), np.ceil((lon_lim[1] + 180) * 20)]
            )
        )

    def download(self, progress_bar: bool = True, cores=None, *args, **kwargs):
        """Download.

            downloads CHIRPS data

        Parameters
        ----------
        progress_bar : TYPE, optional
            will print a waitbar. The default is 1.
        cores : TYPE, optional
            The number of cores used to run the routine. It can be 'False'
                 to avoid using parallel computing routines. The default is None.

        Returns
        -------
        results : TYPE
            DESCRIPTION.
        """
        # Pass variables to parallel function and run
        args = [
            self.path,
            self.temporal_resolution,
            self.xID,
            self.yID,
            self.lon_lim,
            self.lat_lim,
        ]

        if not cores:
            # Create Waitbar
            if progress_bar:
                total_amount = len(self.dates)
                amount = 0
                print_progress_bar(
                    amount,
                    total_amount,
                    prefix="Progress:",
                    suffix="Complete",
                    length=50,
                )

            for date in self.dates:
                self.API(date, args)

                if progress_bar:
                    amount = amount + 1
                    print_progress_bar(
                        amount,
                        total_amount,
                        prefix="Progress:",
                        suffix="Complete",
                        length=50,
                    )
            results = True
        else:
            results = Parallel(n_jobs=cores)(
                delayed(self.API)(date, args) for date in self.dates
            )
        return results

    def API(self, date, args):
        """form the request url abd trigger the request.

        Parameters
        ----------
        date:

        args: [list]
        """
        [path, temp_resolution, xID, yID, lon_lim, latlim] = args

        # Define FTP path to directory
        if temp_resolution.lower() == "daily":
            pathFTP = f"pub/org/chg/products/CHIRPS-2.0/global_daily/tifs/p05/{date.strftime('%Y')}/"
        elif temp_resolution == "monthly":
            pathFTP = "pub/org/chg/products/CHIRPS-2.0/global_monthly/tifs/"
        else:
            raise KeyError("The input temporal_resolution interval is not supported")

        # create all the input name (filename) and output (outfilename, filetif, DiFileEnd) names
        if temp_resolution.lower() == "daily":
            filename = f"{self.globe_fname}.{date.strftime('%Y')}.{date.strftime('%m')}.{date.strftime('%d')}.tif.gz"
            outfilename = os.path.join(
                path,
                f"{self.globe_fname}.{date.strftime('%Y')}.{date.strftime('%m')}.{date.strftime('%d')}.tif",
            )
            DirFileEnd = os.path.join(
                path,
                f"{self.clipped_fname}_mm-day-1_daily_{date.strftime('%Y')}.{date.strftime('%m')}.{date.strftime('%d')}.tif",
            )
        elif temp_resolution == "monthly":
            filename = (
                f"{self.globe_fname}.{date.strftime('%Y')}.{date.strftime('%m')}.tif.gz"
            )
            outfilename = os.path.join(
                path,
                f"{self.globe_fname}.{date.strftime('%Y')}.{date.strftime('%m')}.tif",
            )
            DirFileEnd = os.path.join(
                path,
                f"{self.clipped_fname}_mm-month-1_monthly_{date.strftime('%Y')}.{date.strftime('%m')}.{date.strftime('%d')}.tif",
            )
        else:
            raise KeyError("The input temporal_resolution interval is not supported")

        self.callAPI(pathFTP, path, filename)
        self.post_download(
            path, filename, lon_lim, latlim, xID, yID, outfilename, DirFileEnd
        )

    @staticmethod
    def callAPI(pathFTP: str, path: str, filename: str):
        """send the request to the server.

        RetrieveData method retrieves CHIRPS data for a given date from the
        https://data.chc.ucsb.edu/

        Parameters
        ----------
        filename
        path
        pathFTP


        Raises
        ------
        KeyError
            DESCRIPTION.

        Returns
        -------
        bool
            DESCRIPTION.
        """
        ftp = FTP(CHIRPS.api_url)
        ftp.login()
        # find the document name in this directory
        ftp.cwd(pathFTP)
        listing = []

        # read all the file names in the directory
        ftp.retrlines("LIST", listing.append)

        # download the global rainfall file
        local_filename = os.path.join(path, filename)
        lf = open(local_filename, "wb")
        ftp.retrbinary("RETR " + filename, lf.write, 8192)
        lf.close()

    def post_download(
        self,
        path,
        filename,
        lon_lim,
        latlim,
        xID,
        yID,
        outfilename,
        DirFileEnd,
    ):
        """clip the downloaded data to the extent we want.

        Parameters
        ----------
        path: [str]
            directory where files will be saved
        filename: [str]
            file name
        lon_lim: [list]
        latlim: [list]
        xID: [list]
        yID: [list]
        outfilename: [str]
        DirFileEnd: [str]
        """
        try:
            # unzip the file
            zip_filename = os.path.join(path, filename)
            extractFromGZ(zip_filename, outfilename, delete=True)

            # open tiff file
            src = gdal.Open(outfilename)
            dataset, NoDataValue = Raster.getRasterData(src)

            # clip dataset to the given extent
            data = dataset[yID[0] : yID[1], xID[0] : xID[1]]
            # replace -ve values with -9999
            data[data < 0] = -9999

            # save dataset as geotiff file
            geo = [lon_lim[0], 0.05, 0, latlim[1], 0, -0.05]
            Raster.createRaster(
                DirFileEnd,
                data,
                geo=geo,
                epsg="WGS84",
                nodatavalue=NoDataValue,
            )

            # delete old tif file
            os.remove(outfilename)

        except PermissionError:
            print(
                "The file covering the whole world could not be deleted please delete it after the download ends"
            )
        return True

    def listAttributes(self):
        """Print Attributes List."""

        print("\n")
        print(
            f"Attributes List of: {repr(self.__dict__['name'])} - {self.__class__.__name__} Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                print(str(key) + " : " + repr(self.__dict__[key]))

        print("\n")


class Catalog(AbstractCatalog):
    """CHIRPS data catalog."""

    def __init__(self):
        super().__init__()

    def get_catalog(self):
        """return the catalog."""
        return {
            "Precipitation": {
                "descriptions": "rainfall [mm/temporal_resolution]",
                "units": "mm/temporal_resolution",
                "temporal resolution": ["daily", "monthly"],
                "file name": "rainfall",
                "var_name": "R",
            }
        }

    def get_variable(self, var_name):
        """get the details of a specific variable."""
        return super().get_variable(var_name)



from typing import List, Optional, Union

import ee
import pandas as pd
from ee.featurecollection import FeatureCollection
from ee.geometry import Geometry
from geopandas.geodataframe import GeoDataFrame
from loguru import logger
from shapely.geometry import LineString, Point, Polygon


def createGeometry(
    shapely_geometry: Union[Polygon, Point, LineString],
    epsg: int = 4326,
) -> Geometry:
    """createGeometry.

        create earth engine geometry.

    Parameters
    ----------
    shapely_geometry: [shapely.geometry]
        shapely geometry object [point, polyline, Linestring]
    epsg: [int]
        projection epsg number.

    Returns
    -------
    ee Geometry :
        ee geometry object [Polygon, Point, LineString]
    """
    coords = shapely_geometry.__geo_interface__["coordinates"]
    if shapely_geometry.geom_type == "Polygon":
        return ee.Geometry.Polygon(coords, f"epsg:{epsg}")

    elif shapely_geometry.geom_type in ["Point", "LineString"]:
        if shapely_geometry.geom_type == "LineString":
            raise ValueError("LineStrings not yet implemented.")
        else:
            return ee.Geometry.Point(coords, f"epsg:{epsg}")

    else:
        logger.debug(
            f"The given geometry is neiter of type LineString, Point nor Polygon, "
            f"but {shapely_geometry.geom_type}."
        )


def createFeature(
    gdf: GeoDataFrame, columns: Optional[List[str]] = None
) -> FeatureCollection:
    """createFeature.

        createFeature creates a feature collection from a geodataframe

        collection with the data in a certain column in the geodataframe as a properties dictionary

    Parameters
    ----------
    gdf : [GeoDataFrame]

    columns: [list]
        list of strings for the columns' names

    Returns
    -------
    FeatureCollection : [ee.featurecollection.FeatureCollection]
        feature collection containing the geometry of each row in the given geodataframe
        with the information of one of the given columns as a property.
    """
    try:
        # get the geometry type for all rows
        geotype = [i.geom_type for i in gdf["geometry"]]
        # if any is "MultiPolygon" explode the dataframe to single polygons
        if "MultiPolygon" in geotype:
            # index_parts=True makes the resulted index multi-index if multi-polygon resulted in
            # many different polygons
            gdf = gdf.explode(index_parts=True)

        ee_geom_list = gdf.geometry.apply(lambda geom: createGeometry(geom)).to_list()
        records_df = pd.DataFrame(gdf.drop("geometry", axis=1))
        if columns:
            records_df = records_df[columns]
        records = records_df.to_dict("records")
        if not records:
            ee_feature_list = [ee.Feature(geom) for geom in ee_geom_list]
        else:
            ee_feature_list = [
                ee.Feature(geom, record) for geom, record in zip(ee_geom_list, records)
            ]
        return ee.FeatureCollection(ee_feature_list)

    except Exception as error:
        logger.error(error)
        raise ValueError(error)



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict


class AbstractDataSource(ABC):
    """Bluebrint for all class for different datasources."""

    def __init__(
        self,
        start: str = None,
        end: str = None,
        variables: list = None,
        temporal_resolution: str = "daily",
        lat_lim: list = None,
        lon_lim: list = None,
        fmt: str = "%Y-%m-%d",
        path: str = "",
    ):
        """

        Parameters
        ----------
        temporal_resolution (str, optional):
            [description]. Defaults to 'daily'.
        start (str, optional):
            [description]. Defaults to ''.
        end (str, optional):
            [description]. Defaults to ''.
        path (str, optional):
            Path where you want to save the downloaded data. Defaults to ''.
        variables (list, optional):
            Variable code: VariablesInfo('day').descriptions.keys(). Defaults to [].
        lat_lim (list, optional):
            [ymin, ymax]. Defaults to None.
        lon_lim (list, optional):
            [xmin, xmax]. Defaults to None.
        fmt (str, optional):
            [description]. Defaults to "%Y-%m-%d".
        """
        # initialize connection with ecmwf server
        self.initialize()
        self.temporal_resolution = temporal_resolution
        # TODO: create a function to check if the given variable exists in the catalog
        self.vars = variables

        self.create_grid(lat_lim, lon_lim)
        self.check_input_dates(start, end, temporal_resolution, fmt)
        self.path = Path(path).absolute()
        # Create the directory
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        pass

    @abstractmethod
    def check_input_dates(
        self, start: str, end: str, temporal_resolution: str, fmt: str
    ):
        """Check validity of input dates."""
        pass

    @abstractmethod
    def initialize(self, *args, **kwargs):
        """Initialize connection with the data source server (for non ftp servers)"""
        pass

    @abstractmethod
    def create_grid(self, lat_lim: list, lon_lim: list):
        """create a grid from the lat/lon boundaries."""
        pass

    @abstractmethod
    def download(self):
        """Wrapper over all the given variables."""
        # loop over dates if the downloaded rasters/netcdf are for a specific date out of the required
        # list of dates
        pass

    # @abstractmethod
    def downloadDataset(self):
        """Download single variable/dataset."""
        # used for non ftp servers
        pass

    @abstractmethod
    def API(self):
        """send/recieve request to the dataset server."""
        pass


class AbstractCatalog(ABC):
    """abstrach class for the datasource catalog."""

    def __init__(self):
        self.catalog = self.get_catalog()

    @abstractmethod
    def get_catalog(self):
        """read the catalog of the datasource from disk or retrieve it from server."""
        pass

    @abstractmethod
    def get_variable(self, var_name) -> Dict[str, str]:
        """get the details of a specific variable."""
        return self.catalog.get(var_name)



import os

import pandas as pd
from pandas import DataFrame

from earth2observe.gee import __path__


def getCatalog() -> DataFrame:
    """get_catalog.

        get_catalog retrieves the dataset catalog

    Returns
    -------
    DataFrame
    """
    return pd.read_json(os.path.join(__path__[0], "dataset_catalog.json"))



import datetime as dt

# import ee
from geopandas.geodataframe import GeoDataFrame

from earth2observe.gee.data import getCatalog
from earth2observe.gee.gee import GEE

catalog = getCatalog()
default_date_format = "%Y-%m-%d"


class Dataset(GEE):
    """Dataset."""

    def __init__(
        self,
        dataset_id: str,
        start_date: str,
        end_date: str,
        date_format: str = "%Y-%m-%d",
    ):
        if dataset_id not in catalog["dataset"].tolist():
            raise ValueError(
                f"the given dataset: {dataset_id} does nor exist in the catalog"
            )
        else:
            self.metadata = catalog.loc[catalog["dataset"] == dataset_id, :]
            self.id = id

        self.start_date, self.end_date = self.getDate(
            dataset_id, start_date, end_date, date_format
        )
        # self.catalog = catalog
        self.boundary = None

    @staticmethod
    def getDate(
        dataset_id: str,
        start_date: str = None,
        end_date: str = None,
        date_format: str = default_date_format,
    ):
        """getDate.

            getDate retrieves the start and end date of a dataset

        Parameters
        ----------
        dataset_id: [str]
            dataset id as in the catalog.
        start_date: [str]
            to check it the given start date falls in the available dataset
        end_date: [str]
            to check it the given end date falls in the available dataset
        date_format: [str]
            format of the given dates, Default is YYYY-MM-DD

        Returns
        -------
        start_date: [str]
            beginning of the temporal_resolution series.

        end_date: [str]
            end of the temporal_resolution series.
        """
        data = catalog.loc[catalog["dataset"] == dataset_id, :]

        dataset_start_date = dt.datetime.strptime(
            data["start_date"].values[0], default_date_format
        )
        dataset_end_date = data["end_date"].values[0]
        if dataset_end_date == "Now":
            dataset_end_date = dt.datetime.now().date()

        if not start_date:
            start_date = dt.datetime.strptime(start_date, date_format)
            if start_date < dataset_start_date:
                start_date = dataset_start_date
        else:
            start_date = dataset_start_date

        if not end_date:
            end_date = dt.datetime.strptime(end_date, date_format)
            if end_date > dataset_end_date:
                end_date = dataset_end_date
        else:
            end_date = dataset_end_date

        return start_date, end_date

    def addBoundary(self, gdf: GeoDataFrame):
        """addBoundary.

            addBoundary

        Parameters
        ----------
        gdf
        """
        self.boundary = gdf.copy()

    def filterByRegion(self, gdf: GeoDataFrame = None):
        """filterByRegion.

            filterByRegion

        Parameters
        ----------
        gdf
        """
        if gdf:
            self.addBoundary(gdf)



"""Front end module that runs each data source backend."""
from earth2observe.chirps import CHIRPS
from earth2observe.ecmwf import ECMWF
from earth2observe.s3 import S3

DEFAULT_LONGITUDE_LIMIT = [-180, 180]
DEFAULT_LATITUDE_LIMIT = [-90, 90]


class Earth2Observe:
    """End user class to call all the data source classes abailable in earth2observe."""

    DataSources = {"ecmwf": ECMWF, "chirps": CHIRPS, "amazon-s3": S3}

    def __init__(
        self,
        data_source: str = "chirps",
        temporal_resolution: str = "daily",
        start: str = None,
        end: str = None,
        path: str = "",
        variables: list = None,
        lat_lim: list = None,
        lon_lim: list = None,
        fmt: str = "%Y-%m-%d",
    ):
        """
        Parameters
        ----------
        data_source: [str]
            data source name. the available data sources are ["ecmwf", "chirps", "amazon-s3"].
        temporal_resolution (str, optional):
            temporal resolution. Defaults to 'daily'.
        start (str, optional):
            start date. Defaults to ''.
        end (str, optional):
            end date. Defaults to ''.
        path (str, optional):
            Path where you want to save the downloaded data. Defaults to ''.
        variables (list, optional):
            Variable name.
        lat_lim (list, optional):
            [ymin, ymax]. Defaults to None.
        lon_lim (list, optional):
            [xmin, xmax]. Defaults to None.
        fmt (str, optional):
            date format. Defaults to "%Y-%m-%d".
        """
        if data_source not in self.DataSources:
            raise ValueError(f"{data_source} not supported")

        if lat_lim is None:
            lat_lim = DEFAULT_LATITUDE_LIMIT
        if lon_lim is None:
            lon_lim = DEFAULT_LONGITUDE_LIMIT

        self.datasource = self.DataSources[data_source](
            start=start,
            end=end,
            variables=variables,
            lat_lim=lat_lim,
            lon_lim=lon_lim,
            temporal_resolution=temporal_resolution,
            path=path,
            fmt=fmt,
        )

    def download(self, progress_bar: bool = True, *args, **kwargs):
        self.datasource.download(progress_bar=progress_bar, *args, **kwargs)



"""Amazon S3."""
import datetime as dt
import os
from typing import Dict, List

import boto3
import botocore
import pandas as pd
from botocore import exceptions
from serapeum_utils.utils import print_progress_bar

from earth2observe.abstractdatasource import AbstractCatalog, AbstractDataSource


class S3(AbstractDataSource):
    """Amazon S3 data source."""

    def __init__(
        self,
        temporal_resolution: str = "monthly",
        start: str = None,
        end: str = None,
        path: str = "",
        variables: list = "precipitation_amount_1hour_Accumulation",
        lat_lim: list = None,
        lon_lim: list = None,
        fmt: str = "%Y-%m-%d",
    ):
        """S3.

        Parameters
        ----------
        temporal_resolution (str, optional):
            'daily' or 'monthly'. Defaults to 'daily'.
        start (str, optional):
            [description]. Defaults to ''.
        end (str, optional):
            [description]. Defaults to ''.
        path (str, optional):
            Path where you want to save the downloaded data. Defaults to ''.
        variables (list, optional):
            Variable code: VariablesInfo('day').descriptions.keys(). Defaults to [].
        lat_lim (list, optional):
            [ymin, ymax] (values must be between -50 and 50). Defaults to [].
        lon_lim (list, optional):
            [xmin, xmax] (values must be between -180 and 180). Defaults to [].
        fmt (str, optional):
            [description]. Defaults to "%Y-%m-%d".
        """
        super().__init__(
            start=start,
            end=end,
            variables=variables,
            temporal_resolution=temporal_resolution,
            lat_lim=lat_lim,
            lon_lim=lon_lim,
            fmt=fmt,
            path=path,
        )

    def initialize(self, bucket: str = "era5-pds") -> object:
        """initialize connection with amazon s3 and create a client.

        Parameters
        ----------
        bucket: [str]
            S3 bucket name.

        Returns
        -------
        client: [botocore.client.S3]
            Amazon S3 client
        """
        # AWS access / secret keys required
        # s3 = boto3.resource('s3')
        # bucket = s3.Bucket(era5_bucket)

        # No AWS keys required
        client = boto3.client(
            "s3", config=botocore.client.Config(signature_version=botocore.UNSIGNED)
        )
        self.client = client
        return client

    def create_grid(self, lat_lim: list, lon_lim: list):
        """TODO:"""
        pass

    def check_input_dates(
        self, start: str, end: str, temporal_resolution: str, fmt: str
    ):
        """check validity of input dates.

        Parameters
        ----------
        temporal_resolution: (str, optional)
            [description]. Defaults to 'daily'.
        start: (str, optional)
            [description]. Defaults to ''.
        end: (str, optional)
            [description]. Defaults to ''.
        fmt: (str, optional)
            [description]. Defaults to "%Y-%m-%d".
        """
        self.start = dt.datetime.strptime(start, fmt)
        self.end = dt.datetime.strptime(end, fmt)

        # Set required data for the daily option
        if temporal_resolution == "daily":
            self.dates = pd.date_range(self.start, self.end, freq="D")
        elif temporal_resolution == "monthly":
            self.dates = pd.date_range(self.start, self.end, freq="MS")

    def download(self, progress_bar: bool = True):
        """Download wrapper over all given variables.

        ECMWF method downloads ECMWF daily data for a given variable, temporal_resolution
        interval, and spatial extent.


        Parameters
        ----------
        progress_bar : TYPE, optional
            0 or 1. to display the progress bar
        dataset:[str]
            Default is "interim"

        Returns
        -------
        None.
        """
        catalog = Catalog()
        for var in self.vars:
            var_info = catalog.get_variable(var)
            var_s3_name = var_info.get("bucket_name")
            self.downloadDataset(var_s3_name, progress_bar=progress_bar)

    def downloadDataset(self, var: str, progress_bar: bool = True):
        """Download a climate variable.

                    This function downloads ECMWF six-hourly, daily or monthly data.

        Parameters
        ----------
        var: [str]
            variable detailed information
            >>> {
            >>>     'descriptions': 'Evaporation [m of water]',
            >>>     'units': 'mm',
            >>>     'types': 'flux',
            >>>     'temporal resolution': ['six hours', 'daily', 'monthly'],
            >>>     'file name': 'Evaporation',
            >>> }
        progress_bar: [bool]
            True if you want to display a progress bar.
        """
        if progress_bar:
            total_amount = len(self.dates)
            amount = 0
            print_progress_bar(
                amount,
                total_amount,
                prefix="Progress:",
                suffix="Complete",
                length=50,
            )

        for date in self.dates:
            year = date.strftime("%Y")
            month = date.strftime("%m")
            # file path patterns for remote S3 objects and corresponding local file
            s3_data_key = f"{year}/{month}/data/{var}.nc"
            downloaded_file_dir = (
                f"{self.path}/{year}{month}_{self.temporal_resolution}_{var}.nc"
            )

            self.API(s3_data_key, downloaded_file_dir)

            if progress_bar:
                amount = amount + 1
                print_progress_bar(
                    amount,
                    total_amount,
                    prefix="Progress:",
                    suffix="Complete",
                    length=50,
                )

    def API(self, s3_file_path: str, local_dir_fname: str, bucket: str = "era5-pds"):
        """Download file from s3 bucket.

        Parameters
        ----------
        s3_file_path: str
            the whole path for the file inside the bcket. i.e. "2022/02/main.nc"
        local_dir_fname: [str]
            absolute path for the file name and directory in your local drive.
        bucket: [str]
            bucket name. Default is "era5-pds"

        Returns
        -------
        Download the file to your local drive.
        """
        if not os.path.isfile(local_dir_fname):  # check if file already exists
            print(f"Downloading {s3_file_path} from S3...")
            try:
                self.client.download_file(bucket, s3_file_path, local_dir_fname)
            except exceptions.ClientError:
                print(
                    f"Error while downloading the {s3_file_path} please check the file name"
                )
        else:
            print(f"The file {local_dir_fname} already in your local directory")

    @staticmethod
    def parse_response_metadata(response: Dict[str, str]):
        """parse client response.

        Parameters
        ----------
        response:
        >>> {
        >>>     'RequestId': 'E01V3VYXERE4JV8Z',
        >>>     'HostId': 'Nb33s2L/qsQ/oMb8tyT737ymO5sNnAv+KeKEbhvjILbDPvUi0sDVj6zbuRPh/kpmK5BY6Y3EK/A=',
        >>>     'HTTPStatusCode': 200,
        >>>     'HTTPHeaders': {'x-amz-id-2': 'Nb33s2L/qsQ/oMb8tyT737ymO5sNnAv+KeKEbhvjILbDPvUi0sDVj6zbuRPh/kpmK5BY6Y3EK/A=',
        >>>     'x-amz-request-id': 'E01V3VYXERE4JV8Z',
        >>>     'date': 'Sun, 15 Jan 2023 22:36:28 GMT',
        >>>     'x-amz-bucket-region': 'us-east-1',
        >>>     'content-type': 'application/xml',
        >>>     'transfer-encoding': 'chunked',
        >>>     'server': 'AmazonS3'},
        >>>     'RetryAttempts': 0
        >>> }
        """
        response_meta = response.get("ResponseMetadata")
        keys = []
        if response_meta.get("HTTPStatusCode") == 200:
            contents_list = response.get("Contents")
            if contents_list is None:
                print("No objects are available")  # {date.strftime('%B, %Y')}
            else:
                for obj in contents_list:
                    keys.append(obj.get("Key"))
                print(
                    f"There are {len(keys)} objects available for\n--"
                )  # {date.strftime('%B, %Y')}
                for k in keys:
                    print(k)
        else:
            print("There was an error with your request.")

        return keys


class Catalog(AbstractCatalog):
    """S3 data catalog."""

    def __init__(self, bucket: str = "era5-pds"):
        """

        Parameters
        ----------
        bucket :
        """
        super().__init__()
        self.client = self.initialize(bucket=bucket)

    @staticmethod
    def initialize(bucket: str = "era5-pds") -> object:
        """initialize connection with amazon s3 and create a client.

        Parameters
        ----------
        bucket: [str]
            S3 bucket name.

        Returns
        -------
        client: [botocore.client.S3]
            Amazon S3 client
        """
        # AWS access / secret keys required
        # s3 = boto3.resource('s3')
        # bucket = s3.Bucket(era5_bucket)

        # No AWS keys required
        client = boto3.client(
            "s3", config=botocore.client.Config(signature_version=botocore.UNSIGNED)
        )
        return client

    def get_catalog(self):
        """return the catalog."""
        return {
            "precipitation": {
                "descriptions": "rainfall [mm/temporal_resolution]",
                "units": "mm/temporal_resolution",
                "temporal resolution": ["daily", "monthly"],
                "file name": "rainfall",
                "var_name": "R",
                "bucket_name": "precipitation_amount_1hour_Accumulation",
            }
        }

    def get_variable(self, var_name) -> Dict[str, str]:
        """get the details of a specific variable."""
        return super().get_variable(var_name)

    def get_available_years(self, bucket: str = "era5-pds"):
        """The ERA5 data is chunked into distinct NetCDF files per variable, each containing a month of hourly data. These files are organized in the S3 bucket by year, month, and variable name.

        The data is structured as follows:

        /{year}/{month}/main.nc
                       /data/{var1}.nc
                            /{var2}.nc
                            /{....}.nc
                            /{varN}.nc

        - where year is expressed as four digits (e.g. YYYY) and month as two digits (e.g. MM).

        Parameters
        ----------
        bucket: [str]
            S3 bucket name

        Returns
        -------
        List:
            list of years that have available data.
        """
        paginator = self.client.get_paginator("list_objects")
        result = paginator.paginate(Bucket=bucket, Delimiter="/")
        # for prefix in result.search('CommonPrefixes'):
        #     print(prefix.get('Prefix'))
        years = [i.get("Prefix")[:-1] for i in result.search("CommonPrefixes")]
        return years

    def get_available_data(
        self,
        date: str,
        bucket: str = "era5-pds",
        fmt: str = "%Y-%m-%d",
        absolute_path: bool = False,
    ) -> List[str]:
        """get the available data at a given year.

        - Granule variable structure and metadata attributes are stored in main.nc. This file contains coordinate and
        auxiliary variable data. This file is also annotated using NetCDF CF metadata conventions.

        Parameters
        ----------
        date: [str]
            date i.e. "YYYY-mm-dd"
        bucket: [str]
            The bucket you want to get its available data. Default is 'era5-pds'.
        fmt: [str]
            Date format. Default is "%Y-%m-%d".
        absolute_path: [bool]
            True if you want to get the file names including the whole path inside the bucket.
            Default is False.
            >>> absolute_path = True
            [
                '2022/05/air_pressure_at_mean_sea_level.nc',
                 '2022/05/air_temperature_at_2_metres.nc',
                 '2022/05/air_temperature_at_2_metres_1hour_Maximum.nc',
                 '2022/05/air_temperature_at_2_metres_1hour_Minimum.nc',
                 '2022/05/dew_point_temperature_at_2_metres.nc',
                 '2022/05/eastward_wind_at_100_metres.nc'
             ]
            >>> absolute_path = False
            [
                'air_pressure_at_mean_sea_level.nc',
                'air_temperature_at_2_metres.nc',
                'air_temperature_at_2_metres_1hour_Maximum.nc',
                'air_temperature_at_2_metres_1hour_Minimum.nc',
                'dew_point_temperature_at_2_metres.nc',
                'eastward_wind_at_100_metres.nc'
             ]
        Returns
        -------
        List:
            available data in a list
        """
        date_obj = dt.datetime.strptime(date, fmt)
        # date = dt.date(2022,5,1) # update to desired date
        prefix = date_obj.strftime("%Y/%m/")
        response = self.client.list_objects_v2(Bucket=bucket, Prefix=prefix)
        keys = S3.parse_response_metadata(response)
        if absolute_path:
            available_date = keys
        else:
            available_date = [i.split("/")[-1] for i in keys]
        return available_date



"""Created on Fri Apr  2 06:58:20 2021.

@author: mofarrag
"""
import calendar
import datetime as dt
import os
from typing import Dict

import numpy as np
import pandas as pd
import yaml
from ecmwfapi import ECMWFDataServer
from loguru import logger
from netCDF4 import Dataset
from pyramids.raster import Raster
from serapeum_utils.utils import print_progress_bar

from earth2observe import __path__
from earth2observe.abstractdatasource import AbstractCatalog, AbstractDataSource


class ECMWF(AbstractDataSource):
    """RemoteSensing.

    RemoteSensing class contains methods to download ECMWF data
    """

    temporal_resolution = ["daily", "monthly"]
    spatial_resolution = 0.125

    def __init__(
        self,
        temporal_resolution: str = "daily",
        start: str = None,
        end: str = None,
        path: str = "",
        variables: list = None,
        lat_lim: list = None,
        lon_lim: list = None,
        fmt: str = "%Y-%m-%d",
    ):
        """ECMWF.

        Parameters
        ----------
        temporal_resolution (str, optional):
            [description]. Defaults to 'daily'.
        start (str, optional):
            [description]. Defaults to ''.
        end (str, optional):
            [description]. Defaults to ''.
        path (str, optional):
            Path where you want to save the downloaded data. Defaults to ''.
        variables (list, optional):
            Variable code: VariablesInfo('day').descriptions.keys(). Defaults to [].
        lat_lim (list, optional):
            [ymin, ymax]. Defaults to None.
        lon_lim (list, optional):
            [xmin, xmax]. Defaults to None.
        fmt (str, optional):
            [description]. Defaults to "%Y-%m-%d".
        """
        super().__init__(
            start=start,
            end=end,
            variables=variables,
            temporal_resolution=temporal_resolution,
            lat_lim=lat_lim,
            lon_lim=lon_lim,
            fmt=fmt,
            path=path,
        )

    def check_input_dates(
        self, start: str, end: str, temporal_resolution: str, fmt: str
    ):
        """check validity of input dates.

        Parameters
        ----------
        temporal_resolution: (str, optional)
            [description]. Defaults to 'daily'.
        start: (str, optional)
            [description]. Defaults to ''.
        end: (str, optional)
            [description]. Defaults to ''.
        fmt: (str, optional)
            [description]. Defaults to "%Y-%m-%d".
        """
        self.start = dt.datetime.strptime(start, fmt)
        self.end = dt.datetime.strptime(end, fmt)

        if temporal_resolution == "six_hourly":
            # Set required data for the three hourly option
            self.string1 = "oper"
        # Set required data for the daily option
        elif temporal_resolution == "daily":
            self.dates = pd.date_range(self.start, self.end, freq="D")
        elif temporal_resolution == "monthly":
            self.dates = pd.date_range(self.start, self.end, freq="MS")

        self.date_str = f"{self.start}/to/{self.end}"

    def initialize(self):
        """Initialize connection with ECMWF server."""
        try:
            url = os.environ["ECMWF_API_URL"]
            key = os.environ["ECMWF_API_KEY"]
            email = os.environ["ECMWF_API_EMAIL"]
        except KeyError:
            raise AuthenticationError(
                "Please define the following environment variables to successfully establish a "
                "connection with ecmwf server ECMWF_API_URL, ECMWF_API_KEY, ECMWF_API_EMAIL"
            )

        self.server = ECMWFDataServer(url=url, key=key, email=email)

    def create_grid(self, lat_lim: list, lon_lim: list):
        """Create_grid.

            create grid from the lat/lon boundaries

        Parameters
        ----------
        lat_lim: []
            latitude boundaries
        lon_lim: []
            longitude boundaries
        """
        cell_size = self.spatial_resolution
        # correct latitude and longitude limits
        latlim_corr_one = np.floor(lat_lim[0] / cell_size) * cell_size
        latlim_corr_two = np.ceil(lat_lim[1] / cell_size) * cell_size
        self.latlim_corr = [latlim_corr_one, latlim_corr_two]

        # correct latitude and longitude limits
        lonlim_corr_one = np.floor(lon_lim[0] / cell_size) * cell_size
        lonlim_corr_two = np.ceil(lon_lim[1] / cell_size) * cell_size
        self.lonlim_corr = [lonlim_corr_one, lonlim_corr_two]

    def download(
        self, dataset: str = "interim", progress_bar: bool = True, *args, **kwargs
    ):
        """Download wrapper over all given variables.

        ECMWF method downloads ECMWF daily data for a given variable, temporal_resolution
        interval, and spatial extent.


        Parameters
        ----------
        progress_bar : TYPE, optional
            0 or 1. to display the progress bar
        dataset:[str]
            Default is "interim"

        Returns
        -------
        None.
        """
        # read the datasource catalog
        catalog = Catalog()

        for var in self.vars:
            # Download data
            logger.info(
                f"Download ECMWF {var} data for period {self.start} till {self.end}"
            )
            var_info = catalog.get_variable(var)
            self.downloadDataset(var_info, dataset=dataset, progress_bar=progress_bar)
        # delete the downloaded netcdf
        del_ecmwf_dataset = os.path.join(self.path, "data_interim.nc")
        os.remove(del_ecmwf_dataset)

    def downloadDataset(
        self,
        var_info: Dict[str, str],
        dataset: str = "interim",
        progress_bar: bool = True,
    ):
        """Download a climate variable.

            This function downloads ECMWF six-hourly, daily or monthly data.

        Parameters
        ----------
        var_info: [str]
            variable detailed information
            >>> {
            >>>     'descriptions': 'Evaporation [m of water]',
            >>>     'units': 'mm',
            >>>     'types': 'flux',
            >>>     'temporal resolution': ['six hours', 'daily', 'monthly'],
            >>>     'file name': 'Evaporation',
            >>>     'download type': 2,
            >>>     'number_para': 182,
            >>>     'var_name': 'e',
            >>>     'factors_add': 0,
            >>>     'factors_mul': 1000
            >>> }
        dataset: [str]
            Default is "interm"
        progress_bar: [bool]
            True if you want to display a progress bar.
        """
        # trigger the request to the server
        self.API(var_info, dataset)
        # process the downloaded data
        self.post_download(var_info, self.path, dataset, progress_bar)

    def API(self, var_info, dataset):
        """form the request url abd trigger the request.

        Parameters
        ----------
        dataset: [str]
            dataset name
        var_info: [str]
            variable detailed information
            >>> {
            >>>     'descriptions': 'Evaporation [m of water]',
            >>>     'units': 'mm',
            >>>     'types': 'flux',
            >>>     'temporal resolution': ['six hours', 'daily', 'monthly'],
            >>>     'file name': 'Evaporation',
            >>>     'download type': 2,
            >>>     'number_para': 182,
            >>>     'var_name': 'e',
            >>>     'factors_add': 0,
            >>>     'factors_mul': 1000
            >>> }
        """
        download_type = var_info.get("download type")
        # https://www.ecmwf.int/en/computing/software/ecmwf-web-api
        stream = "oper"  # https://apps.ecmwf.int/codes/grib/format/mars/stream/
        # step is 0, 3, 6, 9, 12
        # time_str is 0, 6, 12, 18
        # levtype: sfc for surface, pl for pressure levels (DEFAULT), pv for potential vorticity level
        # type : 'fc' for forecast field, an for analysis field, ob for Observations, ai for Analysis input,
        #       im for Satellite images, ofb for ODB observation feedback, mfb for Monitoring feedback,
        #       oai for ODB analysis input, wp for Weather parameters (in BUFR format)
        # key words for the API https://confluence.ecmwf.int/display/UDOC/Identification+keywords
        if download_type == 1:
            step = "0"
            time_str = "00:00:00/06:00:00/12:00:00/18:00:00"
            levtype = "sfc"  # surface
            type_str = "an"  # analysis field

        if download_type == 2:
            step = "12"
            time_str = "00:00:00/12:00:00"
            levtype = "sfc"  # surface
            type_str = "fc"  # forecast field

        if download_type == 3:
            step = "0"
            time_str = "00:00:00/06:00:00/12:00:00/18:00:00"
            levtype = "pl"  # pressure levels (DEFAULT)
            type_str = "an"  # analysis field

        parameter_number = var_info.get("number_para")

        param = f"{parameter_number}.128"
        grid = f"{self.spatial_resolution}/{self.spatial_resolution}"
        class_str = "ei"
        # N, W, S, E
        area_str = f"{self.latlim_corr[1]}/{self.lonlim_corr[0]}/{self.latlim_corr[0]}/{self.lonlim_corr[1]}"

        # Download data by using the ECMWF API
        logger.info("Use API ECMWF to collect the data, please wait")
        self.callAPI(
            self.server,
            self.path,
            download_type,
            stream,
            levtype,
            param,
            step,
            grid,
            time_str,
            self.date_str,
            type_str,
            class_str,
            area_str,
            dataset=dataset,
        )

    @staticmethod
    def callAPI(
        server,
        output_folder: str,
        download_type: str,
        stream: str,
        levtype: str,
        param: str,
        step: str,
        grid: str,
        time_str: str,
        date_str: str,
        type_str: str,
        class_str: str,
        area_str: str,
        dataset: str = "interim",
    ):
        """send the request to the server.

        Parameters
        ----------
        output_folder: [str]
            directory where files will be saved
        download_type: [int]
        stream
        levtype
        param
        step
        grid
        time_str
        date_str
        type_str
        class_str
        area_str
        dataset
        """
        if download_type == 1 or download_type == 2:
            server.retrieve(
                {
                    "stream": stream,
                    "levtype": levtype,
                    "param": param,
                    "dataset": dataset,
                    "step": step,
                    "grid": grid,
                    "time": time_str,
                    "date": date_str,
                    "type": type_str,  # http://apps.ecmwf.int/codes/grib/format/mars/type/
                    "class": class_str,  # http://apps.ecmwf.int/codes/grib/format/mars/class/
                    "area": area_str,
                    "format": "netcdf",
                    "target": f"{output_folder}/data_{dataset}.nc",
                }
            )

        if download_type == 3:
            server.retrieve(
                {
                    "levelist": "1000",
                    "stream": stream,
                    "levtype": levtype,
                    "param": param,
                    "dataset": dataset,
                    "step": step,
                    "grid": grid,
                    "time": time_str,
                    "date": date_str,
                    "type": type_str,  # http://apps.ecmwf.int/codes/grib/format/mars/type/
                    "class": class_str,  # http://apps.ecmwf.int/codes/grib/format/mars/class/
                    "area": area_str,
                    "format": "netcdf",
                    "target": f"{output_folder}/data_{dataset}.nc",
                }
            )

    def post_download(
        self, var_info: Dict[str, str], out_dir, dataset: str, progress_bar: bool = True
    ):
        """clip the downloaded data to the extent we want.

        Parameters
        ----------
        var_info: [str]
            variable detailed information
            >>> {
            >>>     'descriptions': 'Evaporation [m of water]',
            >>>     'units': 'mm',
            >>>     'types': 'flux',
            >>>     'temporal resolution': ['six hours', 'daily', 'monthly'],
            >>>     'file name': 'Evaporation',
            >>>     'download type': 2,
            >>>     'number_para': 182,
            >>>     'var_name': 'e',
            >>>     'factors_add': 0,
            >>>     'factors_mul': 1000
            >>> }
        out_dir: [str]
            root directory for where the files will be saved.
        dataset: [str]
            dataset name. Default is interm
        progress_bar: [bool]
            True to display a progress bar
        """
        # Open the downloaded data
        NC_filename = os.path.join(self.path, f"data_{dataset}.nc")
        fh = Dataset(NC_filename, mode="r")

        # Get the NC variable parameter
        parameter_var = var_info.get("var_name")
        Var_unit = var_info.get("units")
        factors_add = var_info.get("factors_add")
        factors_mul = var_info.get("factors_mul")

        # Open the NC data
        Data = fh.variables[parameter_var][:]
        Data_time = fh.variables["time"][:]
        lons = fh.variables["longitude"][:]
        lats = fh.variables["latitude"][:]

        # Define the georeference information
        geo_four = np.nanmax(lats)
        geo_one = np.nanmin(lons)
        geo = tuple(
            [
                geo_one,
                self.spatial_resolution,
                0.0,
                geo_four,
                0.0,
                -1 * self.spatial_resolution,
            ]
        )

        # Create Waitbar
        if progress_bar:
            total_amount = len(self.dates)
            amount = 0
            print_progress_bar(
                amount, total_amount, prefix="Progress:", suffix="Complete", length=50
            )

        for date in self.dates:

            # Define the year, month and day
            year = date.year
            month = date.month
            day = date.day

            # Hours since 1900-01-01
            start = dt.datetime(year=1900, month=1, day=1)
            end = dt.datetime(year, month, day)
            diff = end - start
            hours_from_start_begin = diff.total_seconds() / 60 / 60

            Date_good = np.zeros(len(Data_time))

            if self.temporal_resolution == "daily":
                days_later = 1
            if self.temporal_resolution == "monthly":
                days_later = calendar.monthrange(year, month)[1]

            Date_good[
                np.logical_and(
                    Data_time >= hours_from_start_begin,
                    Data_time < (hours_from_start_begin + 24 * days_later),
                )
            ] = 1

            Data_one = np.zeros(
                [int(np.sum(Date_good)), int(np.size(Data, 1)), int(np.size(Data, 2))]
            )
            Data_one = Data[np.int_(Date_good) == 1, :, :]

            # convert the values to the units we want
            Data_end = factors_mul * np.nanmean(Data_one, 0) + factors_add

            if var_info.get("types") == "flux":
                Data_end = Data_end * days_later

            var_output_name = var_info.get("file name")

            # Define the out name
            name_out = os.path.join(
                out_dir,
                f"{var_output_name}_ECMWF_ERA-Interim_{Var_unit}_{self.temporal_resolution}_{year}.{month}.{day}.tif",
            )

            # Create Tiff files
            # Raster.Save_as_tiff(name_out, Data_end, geo, "WGS84")
            Raster.createRaster(path=name_out, arr=Data_end, geo=geo, epsg="WGS84")

            if progress_bar:
                amount = amount + 1
                print_progress_bar(
                    amount,
                    total_amount,
                    prefix="Progress:",
                    suffix="Complete",
                    length=50,
                )

        fh.close()


class Catalog(AbstractCatalog):
    """ECMWF data catalog This class contains the information about the ECMWF variables http://rda.ucar.edu/cgi-bin/transform?xml=/metadata/ParameterTables/WMO_GRIB1.98-0.128.xml&view=gribdoc."""

    def __init__(self):
        # get the catalog
        super().__init__()
        pass

    def get_catalog(self):
        """readthe data catalog from disk."""
        with open(f"{__path__[0]}/ecmwf_data_catalog.yaml", "r") as stream:
            catalog = yaml.safe_load(stream)
        return catalog

    def get_variable(self, var_name):
        """retrieve a variable form the datasource catalog."""
        return super().get_variable(var_name)


class AuthenticationError(Exception):
    """Failed to establish connection with ECMWF server."""

    pass
